package redactableBlockchain;

import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Observable;
import java.util.Random;

import org.apache.commons.collections4.OrderedMapIterator;
import org.apache.commons.collections4.map.LinkedMap;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import redactableBlockchain.data.Block;
import redactableBlockchain.data.Chain;
import redactableBlockchain.data.VirtualBlock;
import redactableBlockchain.data.IBlock;
import redactableBlockchain.data.CorrectionBlock;
import redactableBlockchain.data.StandardBlock;
import redactableBlockchain.data.transaction.ElectionTransaction;
import redactableBlockchain.data.transaction.PaymentTransaction;
import redactableBlockchain.data.transaction.Transaction;
import redactableBlockchain.data.transaction.VoteQuestion;
import redactableBlockchain.data.transaction.VotingTransaction;
import redactableBlockchain.exception.BlockInvalidException;
import redactableBlockchain.exception.VerificationException;
import redactableBlockchain.util.Constants;
import redactableBlockchain.visualization.CreatePicture;


public class UnoptimizedRedactableBlockChainImpl extends Observable implements IBlockChain
{
	private static final Logger logger = LoggerFactory.getLogger(UnoptimizedRedactableBlockChainImpl.class);
	public static final int MEDIAN_BLOCKS = 11;
	//	public static final long COINBASE_MATURITY = 100;
	//
	private String hashOfLastBlockInLongestChain = "";

	private HashSet<IBlockChainListener> listeners = new HashSet<>();

	private HashMap<String, IBlock> normalBlocks = new HashMap<>();
	private HashMap<String, HashSet<IBlock>> normalBlocksByPreviousHash = new HashMap<>();

	private HashMap<String, CorrectionBlock> modifyBlocks = new HashMap<>();

	private HashMap<String, HashSet<String>> linkExceptions = new HashMap<>();

	private HashMap<String, ElectionTransaction> electionMap = new HashMap<>();

	Chain longestChain = new Chain();
	private Wallet wallet = Constants.getAliceWallet();
	private PublicKey coinBaseAddress = wallet.getPublicKey();//TODO
	private boolean print = false;
	private boolean createDot = false;
	//	private boolean mineNormalBlock = true;
	public UnoptimizedRedactableBlockChainImpl(boolean print) throws VerificationException, NoSuchAlgorithmException, InvalidKeySpecException
	{
		this.print = print;
	}
	public UnoptimizedRedactableBlockChainImpl(boolean print, boolean createDot) throws VerificationException, NoSuchAlgorithmException, InvalidKeySpecException
	{
		this.print = print;
		this.createDot = createDot;
	}

	public void addListener(IBlockChainListener listener)
	{
		this.listeners.add(listener);
	}

	private void createNewBlockTemplateAndNotifyNewBlockToMine() throws NoSuchAlgorithmException, InvalidKeySpecException {
		IBlock b = getBlockTemplate();
		if(b instanceof StandardBlock)
		{
			randomlyCreateElectionsOrVotes((StandardBlock) b);
		}

		notifyNewBlockToMine(b);
	}

	private StandardBlock randomlyCreateElectionsOrVotes(StandardBlock b) throws NoSuchAlgorithmException, InvalidKeySpecException
	{
		if(longestChain.size() <=2)
			return b;
		Random rnd = new Random();
		String hashOfElectionTX = longestChain.getActiveVoting();

		//randomly vote if voting is active
		if(hashOfElectionTX != null)
		{
			int val = rnd.nextInt(11);
			VotingTransaction tx;
			if(val <= 8)
			{
				tx = new VotingTransaction(wallet.getPublicKeyAsString(), hashOfElectionTX, true);
			}
			else
			{
				tx = new VotingTransaction(wallet.getPublicKeyAsString(), hashOfElectionTX, false);
			}
			try {
				tx.sign(wallet.privateKey);
				b.addTransaction(tx);
			} catch (VerificationException e) {
				e.printStackTrace();
			}
		}
		else
		{
			int val = rnd.nextInt(11);
			//randomly start a voting 
			if(val <= 1)
			{
				VoteQuestion voteQuestion = createVoteQuestionToDeleteARandomBlock();
				try {
					ElectionTransaction tx = new ElectionTransaction(coinBaseAddress, voteQuestion );
					tx.sign(wallet.privateKey);
					b.addTransaction(tx);
				} catch (VerificationException | NoSuchAlgorithmException | InvalidKeySpecException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
		return b;
	}

	private VoteQuestion createVoteQuestionToDeleteARandomBlock() {
		Random rnd = new Random();
		int index = -1;
		do
		{
			index = 1 + rnd.nextInt(longestChain.size()-2);//don't select first or last block
			String key = longestChain.get(index);
			IBlock b = longestChain.get(key);
			ElectionTransaction tx = blockContainsElection(b);
			VotingTransaction vote = blockContainsVote(b);
			if(tx == null && vote == null && !(b instanceof VirtualBlock))
			{
				break;
			}
			else
				index = -1;

		}while(index == -1);

		String key = longestChain.get(index);
		StandardBlock blockToReplace = (StandardBlock) longestChain.get(key);
		String hashOfPreviousBlock = blockToReplace.getPreviousBlockHash();

		key = longestChain.get(index+1);//subsequent block

		String hashOfSubsequentBlock = longestChain.get(key).calculateHash();
		ArrayList<Transaction> tx = blockToReplace.getTransactions();//use same TX, just show that we can replace the block
		VoteQuestion voteQuestion = new VoteQuestion(blockToReplace.calculateHash(), hashOfPreviousBlock, hashOfSubsequentBlock, blockToReplace.getCreationTime(), tx);
		return voteQuestion;
	}

	@Override
	public IBlock getBlockTemplate() {
		//		NormalBlock lastNormalBlock = getLastBlock();
		//		ModifyBlock lastModifyBlock = getLastModifyBlock();
		Block b = getLastBlock();
		if(b instanceof VirtualBlock)
		{
			//cannot be the last block 
			return null;
		}
		String previousHash = b.getHash();
		IBlock blockToMine = null;
		try {
			//check if normalblock or redaction block
			if(longestChain.isCreateModifyBlock() == true)
			{
				ElectionTransaction tx = longestChain.getEndedVoting();
				electionMap.put(tx.calculateHash(), tx);
				String previousHashChain2 = ((StandardBlock)b).getLastRedactionBlockHash();
				blockToMine = new CorrectionBlock(tx.getVoteQuestion().getHashOfPreviousBlock(), tx.getVoteQuestion().getHashOfSubsequentBlock(), previousHash, previousHashChain2, tx.calculateHash(), createCoinBaseTransaction());
			}
			else
			{
				String lastRedactionBlockHash = longestChain.getNextRedactionBlockHash();
				blockToMine = new StandardBlock(previousHash, lastRedactionBlockHash, createCoinBaseTransaction());
			}
		} catch (NoSuchAlgorithmException e) {
			e.printStackTrace();
		} catch (InvalidKeySpecException e) {
			e.printStackTrace();
		}
		return blockToMine;
	}

	private PaymentTransaction createCoinBaseTransaction() throws NoSuchAlgorithmException, InvalidKeySpecException
	{
		return Constants.createCoinBaseTransaction(coinBaseAddress);
	}

	private void notifyAddedBlock(StandardBlock block) {

		for(IBlockChainListener listener : listeners)
		{
			listener.notifyAddedBlock(block);
		}
		if(createDot) {
			try {
				CreatePicture.createPNG(normalBlocks, modifyBlocks, "test_",longestChain);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	private void notifyAddedBlock(CorrectionBlock block) {
		for(IBlockChainListener listener : listeners)
		{
			listener.notifyAddedRedactionBlock(block);
		}
		if(createDot) {
			try {
				CreatePicture.createPNG(normalBlocks, modifyBlocks, "test_",longestChain);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	private void notifyMissingBlock(String hash) {
		if(hash.equalsIgnoreCase("0000000000000000000000000000000000000000000000000000000000000000"))
			return;
		if(hash.equalsIgnoreCase(Constants.getGenesisBlock().calculateHash()))
			return;
		for(IBlockChainListener listener : listeners)
		{
			listener.notifyMissingBlock(hash);
		}
	}

	private void notifyNewLongestChain(StandardBlock block) {
		for(IBlockChainListener listener : listeners)
		{
			listener.notifyNewLongestChain(block);
		}
	}

	private void notifyNewBlockToMine(IBlock block) {
		for(IBlockChainListener listener : listeners)
		{
			listener.notifyNewBlockToMine(block);
		}
	}

	/**
	 * Calculate the median of the (some number of) blocks starting at the given block.
	 */
	//	private long getMedianTimestamp(BlockChainLink link)
	//	{
	//		if ( link == null )
	//			return 0;
	//		IBlock block = link.getBlock();
	//		return getMedianTimestamp(block);
	//	}

	//	private long getMedianTimestamp(RedactableBlockChainLink link)
	//	{
	//		if ( link == null )
	//			return 0;
	//		IBlock block = link.getBlock();
	//		return getMedianTimestamp(block);
	//	}
	private long getMedianTimestamp(IBlock block)
	{
		List<Long> times = new LinkedList<>();
		for ( int i=0; (block!=null) && (i<MEDIAN_BLOCKS); i++ )
		{
			times.add(block.getCreationTime());
			block=getBlock(block.getPreviousBlockHash());
		}
		Collections.sort(times);
		return times.get(times.size()/2);
	}

	@Override
	public StandardBlock getGenesisBlock() throws NoSuchAlgorithmException, InvalidKeySpecException
	{
		return Constants.getGenesisBlock();
	}

	@Override
	public Block getLastBlock() {
		return (Block) longestChain.get(longestChain.lastKey());
	}

	@Override
	public IBlock getBlock(String hash) {
		IBlock b = normalBlocks.get(hash);
		if(b == null)
			b = modifyBlocks.get(hash);
		return b;
	}

	@Override
	public synchronized void addBlock(Block block) throws VerificationException, NoSuchAlgorithmException, InvalidKeySpecException {
		//System.out.println("adding block:"+block.getHash());
		if(!(block instanceof StandardBlock) && !(block instanceof VirtualBlock))
		{
			return;
		}
		if(normalBlocks.containsKey(block.getHash()))
		{
			if(block instanceof StandardBlock && normalBlocks.get(block.getHash()) instanceof StandardBlock)
			//block already known
			//			return;//cant do that because of fakeblocks which have the same hash
				return;
		}
		boolean blockNew = addNormalBlock(block);
		if(blockNew != true)
		{
			return;
		}

		//		ElectionTransaction election = blockContainsElection(block);
		//		if(election != null)
		//			electionMap.put(election.calculateHash(), election);
		Chain longestChain = buildAllPossibleChains();
		IBlock b = longestChain.get(longestChain.lastKey());
		String hash = b.calculateHash();
		if(!hash.equalsIgnoreCase(hashOfLastBlockInLongestChain) || longestChain.getNextBlockMustReferenceModifyBlock() != null)
		{
			hashOfLastBlockInLongestChain = hash;
			createNewBlockTemplateAndNotifyNewBlockToMine();
		}
		if(block instanceof StandardBlock)
			notifyAddedBlock((StandardBlock)block);
		if(!normalBlocks.containsKey(block.getPreviousBlockHash()))
			notifyMissingBlock(block.getPreviousBlockHash());
		if(block instanceof StandardBlock)
			if(!modifyBlocks.containsKey(((StandardBlock)block).getLastRedactionBlockHash()))//TODO and its not genesis
				notifyMissingBlock(((StandardBlock)block).getLastRedactionBlockHash());
	}

	private boolean addNormalBlock(IBlock block) {
		String hash = block.calculateHash();
		normalBlocks.put(hash, block);
		HashSet<IBlock> set = null;
		String previoushash = block.getPreviousBlockHash();
		if(normalBlocksByPreviousHash.containsKey(previoushash))
		{
			set = normalBlocksByPreviousHash.get(previoushash);
		}
		if(set == null)
		{
			set = new HashSet<>();
			normalBlocksByPreviousHash.put(previoushash, set);
		}
		return set.add(block);
	}

	@Override
	public synchronized void addModifyBlock(CorrectionBlock block) throws VerificationException, NoSuchAlgorithmException, InvalidKeySpecException {
		if(modifyBlocks.containsKey(block.getHash()))
		{
			//block already known
						return;
		}
		modifyBlocks.put(block.getHash(), block);
		//add FakeBlock to list
		VirtualBlock fakeBlock = createFakeBlock(block);
		//add fakeblock to block list
		addBlock(fakeBlock);
		buildAllPossibleChains();

		notifyAddedBlock(block);
	}

	private VirtualBlock createFakeBlock(CorrectionBlock modifyBlock) {
		//		ElectionTransaction tx = electionMap.get(modifyBlock.getElectionHash());//TODO this information may not aavailable yet...
		//		VoteQuestion vote = tx.getVoteQuestion();
		
//		VoteQuestion votequestion = modifyBlock.getVoteQuestion();
		ElectionTransaction electionTX = electionMap.get(modifyBlock.getElectionHash());
		VoteQuestion votequestion = electionTX.getVoteQuestion();
		
		//ArrayList<Transaction> transactions = new ArrayList(normalBlocks.get(votequestion.getHashOfBlockToReplace()).getTransactions());//TODO this information may not available anymore...
		ArrayList<Transaction> transactions = new ArrayList();//TODO
		VirtualBlock b = new VirtualBlock(votequestion.getHashOfBlockToReplace(), votequestion.getHashOfPreviousBlock(), modifyBlock.calculateHash(), transactions , votequestion.getTimeStamp());
		return b;
	}
	private synchronized Chain buildAllPossibleChains() {
		HashSet<Chain> myChains = new HashSet<>();
		Chain myChain1 = new Chain();
		myChain1.put(Constants.getGenesisBlock().getHash(), Constants.getGenesisBlock());
		myChains.add(myChain1);
		boolean loop = true;
		if(normalBlocks.size() > 2) {
			int af =3;
		}
		while(loop)
		{
			Iterator<Chain> it = myChains.iterator();
			long found = 0;
			HashSet<Chain> myTmpChains = new HashSet<>();
			while (it.hasNext())
			{
				Chain myChain = it.next();
				String lastKey = myChain.lastKey();
				//get all blocks that reference our last block
				HashSet<IBlock> blocks = normalBlocksByPreviousHash.get(lastKey);
				if(blocks != null)
				{
					for(IBlock block : blocks)
					{
						Chain myChainTmp = new Chain(myChain);
						myChainTmp.put(block.calculateHash(), block);
						myTmpChains.add(myChainTmp);
					}
					//				TODO think
					if(blocks.size() > 0)
					{
						it.remove();
						//myChains.remove(myChain);//remove shorter chain => is included in longer ones
						found+= blocks.size();
					}
				}
			}
			myChains.addAll(myTmpChains);
			if(found == 0)
				loop = false;
		}
		//check which chains are valid and which one is the longest
		//it can also be that only a parts of a chain is valid...
		Iterator<Chain> it = myChains.iterator();
		long found = 0;
		Chain longest = new Chain();

		while (it.hasNext())
		{
			Chain myChain = it.next();
			checkIfChainIsValid(myChain);
			//			System.out.println(myChain.size() +" "+myChain);
			if(getSize(myChain) > getSize(longest))
				longest = myChain;
			//			checkIfRedactionsAreOk(myChain);
		}
		longestChain = longest;
		//System.out.println("Longest found chain: "+longest.size()+", blocks:"+normalBlocks.size());
		if(print)
			longest.printChain();
		return longest;
	}

	private int getSize(Chain chain) {
		int redactions = 0;
		OrderedMapIterator<String, IBlock> it = chain.mapIterator();
		while(it.hasNext())
		{
			String key = it.next();
			Block b = (Block) chain.get(key);
			if(b instanceof VirtualBlock)
				redactions++;
		}	
		// TODO calc real difficulty ....
		return chain.size()+redactions ;
	}

	//<<<<>>>>
	private void checkIfRedactionsAreOk(LinkedMap<String, IBlock> myChain) {
		String lastRedactionBlock = "0000000000000000000000000000000000000000000000000000000000000000";
		OrderedMapIterator<String, IBlock> it = myChain.mapIterator();
		int blockHeight = -1;
		boolean removeBlocks = false;
		int votingActive = -1;
		while(it.hasNext())
		{
			String key = it.next();
			IBlock b = myChain.get(key);

			blockHeight++;
			if(blockHeight == 0 && b instanceof StandardBlock && ((StandardBlock)b).getLastRedactionBlockHash() != "0000000000000000000000000000000000000000000000000000000000000000")
				removeBlocks = true;
			else
				lastRedactionBlock = b.calculateHash();
			if(blockHeight > 0)
			{
				if(b instanceof StandardBlock)
				{
					//					if(nicht voting ende)
					//					{
					//						
					//					}
					//					else
					//					{
					//						if(((NormalBlock)b).getLastRedactionBlockHash() != lastRedactionBlock)
					//							removeBlocks = true;
					//					}
				}else if(b instanceof CorrectionBlock)
				{
					System.err.println("Should never happen...");
				}else if(b instanceof VirtualBlock)
				{
					//check if election ended and was won
					//check if new redaction block is linked

				}

			}
			if(removeBlocks)
			{
				it.remove();
				continue;
			}
		}
	}

	private void checkIfChainIsValid(Chain myChain) {
		// TODO 
		//		int election = -1;

		OrderedMapIterator<String, IBlock> it = myChain.mapIterator();
		if(myChain.size() == 10) {
			int hha=3;
		}
		//check if distinct blocks are valid
		boolean removeBlocks = false;
		while(it.hasNext())
		{
			String key = it.next();
			IBlock b = myChain.get(key);
			try {
				checkBlock(b);
			} catch (BlockInvalidException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				removeBlocks = true;
				//todo remove 
			}
			if(removeBlocks)
				it.remove();
		}
		//check if voting rules are complied with
		it = myChain.mapIterator();
		myChain.setVotingActive(-1);
		myChain.setActiveElection(null);
		myChain.setInFavorOfElection(0);
		//		int votingActive = -1;
		//		int inFavorOfElection = 0;

		removeBlocks = false;
		//		myChain.setChainMustIncludeFakeBlockWithHash(null);
		//		boolean nextBlockMustBeAFakeBlock = false; 
		while(it.hasNext())
		{
			String key = it.next();
			if(removeBlocks)
			{
				it.remove();
				continue;
			}
			IBlock b = myChain.get(key);

			if(myChain.getVotingActive() != -1)
				myChain.setVotingActive(myChain.getVotingActive()+1);
			ElectionTransaction election = blockContainsElection(b);
			if(election != null)
			{
				if(myChain.getVotingActive() != -1)
				{
					//remove this and all later blocks
					removeBlocks = true;
				} else
				{
					//					votingActive++;
					//voting starts next block
					myChain.setVotingActive(0);
					myChain.setActiveElection(election);
				}
			}
			if(removeBlocks)
			{
				it.remove();
				continue;
			}
			VotingTransaction vote = blockContainsVote(b);
			if(vote != null)
			{
				if(myChain.getVotingActive() == -1 || myChain.getVotingActive() >Constants.votingDuration)
				{
					removeBlocks = true;//remove this and all later blocks
				}
				else
				{
					if(vote.isApproval())
					{
						myChain.setInFavorOfElection(myChain.getInFavorOfElection()+1);
					}
				}

			}

			//			if(myChain.getChainMustIncludeFakeBlockWithHash() != null)
			//			{//wrong => must be a modify block, but the fakeblock is ealrier in the chain ....
			////				if(!(b instanceof FakeBlock))
			////					removeBlocks = true;//remove this and all later blocks
			//				IBlock bb= myChain.get(myChain.getChainMustIncludeFakeBlockWithHash());
			//				if(bb == null)//block is not in chain
			//					removeBlocks = true;
			//				else if(!(bb instanceof FakeBlock))//block is not a fakeblock => likely old chain without redaction applied
			//					removeBlocks = true;
			//				myChain.setNextBlockMustReferenceADifferentModifyBlock(true);
			//				myChain.setChainMustIncludeFakeBlockWithHash(null);
			//				myChain.setVotingActive(-1);
			//			}
			//			else
			////			{
			////				if((b instanceof FakeBlock))
			////					removeBlocks = true;//remove this and all later blocks
			////			}
			if(removeBlocks)
			{
				it.remove();
				continue;
			}
			if(myChain.getNextBlockMustReferenceModifyBlock() != null)
			{
				//TODO
			}
			if(myChain.getVotingActive() == Constants.votingDuration)//+1 TODO
			{
				//voting ended 
				myChain.setVotingActive(-1);
				if(myChain.getInFavorOfElection() > (Constants.votingDuration/2))
				{
					//election won ..next block should be a redaction block
					String hash = myChain.getActiveElection().getVoteQuestion().getHashOfBlockToReplace();
					IBlock bb= myChain.get(hash);
					if(bb == null)//block is not in chain => create redaction block
					{
						removeBlocks = true;
						//						myChain.setCreateModifyBlock(true);
					}
					else //check if block is fakeblock, if so redaction block was already created
					{
						if((bb instanceof VirtualBlock))//block is not a fakeblock => likely old chain without redaction applied
						{
							//							removeBlocks = true;
							myChain.setCreateModifyBlock(false);
							myChain.setNextBlockMustReferenceModifyBlock(((VirtualBlock) bb).getLastRedactionBlockHash());
						}
						else if(bb instanceof StandardBlock)
						{
							myChain.setCreateModifyBlock(true);
							myChain.setNextBlockMustReferenceModifyBlock(null);
						}
					}
					//						removeBlocks = true;
					//					else if(!(bb instanceof FakeBlock))//block is not a fakeblock => likely old chain without redaction applied
					//						removeBlocks = true;

					//					myChain.setChainMustIncludeFakeBlockWithHash(null);
					//					myChain.setChainMustIncludeFakeBlockWithHash(myChain.getActiveElection().getVoteQuestion().getHashOfBlockToReplace());
				}
				else
				{
					myChain.setCreateModifyBlock(false);
					myChain.setNextBlockMustReferenceModifyBlock(null);
					//election lost ..no new redaction block
					//					myChain.setChainMustIncludeFakeBlockWithHash(null);
				}
				myChain.setActiveElection(null);
			}
			if(removeBlocks)
			{
				it.remove();
				continue;
			}
		}

	}

	private VotingTransaction blockContainsVote(IBlock b) {
		int i = 0;
		for(Transaction tx : b.getTransactions())
		{
			i++;
			if(tx instanceof VotingTransaction)
			{
				if(i != 2)
					System.err.println("Vote TX should be second !?");
				return (VotingTransaction) tx;
			}
		}
		return null;
	}

	private ElectionTransaction blockContainsElection(IBlock b) {
		for(Transaction tx : b.getTransactions())
		{
			if(tx instanceof ElectionTransaction)
			{
				return (ElectionTransaction) tx;
			}
		}
		return null;
	}

	private void checkBlock(IBlock b) throws BlockInvalidException {
		b.validate();

	}

	public void print()
	{
		longestChain.printChain();
	}

	public Chain getLongestChain()
	{
		return longestChain;
	}
}