package at.tugraz.iaik.p2p.pol

import at.tugraz.iaik.p2p.commons.*
import at.tugraz.iaik.p2p.skad.ex.CodecException
import at.tugraz.iaik.p2p.skad.id.ECID
import at.tugraz.iaik.p2p.skad.id.readIdentifier
import at.tugraz.iaik.p2p.skad.id.writeIdentifier
import com.codahale.shamir.Scheme
import kotlinx.io.errors.IOException
import org.jeasy.rules.api.Facts
import org.jeasy.rules.core.DefaultRulesEngine
import org.jeasy.rules.mvel.MVELRuleFactory
import org.jeasy.rules.support.YamlRuleDefinitionReader
import java.io.*
import java.nio.ByteBuffer
import java.security.AlgorithmParameters
import java.security.interfaces.ECPrivateKey
import javax.crypto.Cipher
import javax.crypto.KeyGenerator
import javax.crypto.SecretKey

//Sample policy template that expects the evalautor to set the sender of a request
//This means that the sender's identity is available for evaluating policies
class IdBasedPolicy(val rules: String) {
    val facts = Facts()
    fun setSender(id: ECID) {
        facts.put("sender", id.toString())
        facts.put("accessGranted", false)
    }

    val ruleSet = MVELRuleFactory(YamlRuleDefinitionReader()).createRules(StringReader(rules))
    val ruleEngine = DefaultRulesEngine()
    @Throws(Throwable::class)
    fun evaluate() {
        ruleEngine.fire(ruleSet, facts)
        if (!facts.get<Boolean>("accessGranted")) {
            throw Throwable(
                message = ruleSet.joinToString(
                    separator = "\n",
                    prefix = "\n---\n",
                    postfix = "\n---\n"
                ) { it.description })
        }
    }

    fun toByteArray(): ByteArray {
        val baos = ByteArrayOutputStream(4096)
        val out = DataOutputStream(baos)
        val toWrite = rules.toByteArray()
        out.writeByteArray(toWrite)
        out.close()
        return baos.toByteArray()
    }

    companion object {
        fun fromByteArray(arr: ByteArray): IdBasedPolicy {
            val inp = DataInputStream(ByteArrayInputStream(arr))
            return IdBasedPolicy(String(inp.readByteArray()))
        }
    }
}

data class SignedPolicy(val policy: IdBasedPolicy, val signature: DetachedSignature)

fun DataOutput.writeSignedPolicy(policy: SignedPolicy) {
    writeByteArray(policy.policy.toByteArray())
    writeByteArray(policy.signature)
}

fun DataInput.readSignedPolicy() = SignedPolicy(IdBasedPolicy.fromByteArray(readByteArray()), readByteArray())

//associates shared secret data rpotection with the iphertext it protexts
data class ProtectedData(
    val dataProtection: SharedSecretDataProtection,
    val ciphertext: ByteArray
)

fun DataOutput.writeProtectedData(protectedData: ProtectedData) {
    writeSharedSecretDataProtection(protectedData.dataProtection)
    writeByteArray(protectedData.ciphertext)
}

fun DataInput.readProtectedData() = ProtectedData(readSharedSecretDataProtection(), readByteArray())


//Key share container associating the actual share of a key with the evaluator's ID it is intended for, als well as a signature
data class Share(val evaluator: ECID, val fragment: KeyShare, val signature: DetachedSignature)
typealias KeyShare = ByteArray

fun DataOutput.writeShare(share: Share) {
    writeIdentifier(share.evaluator)
    writeByteArray(share.fragment)
    writeByteArray(share.signature)
}

fun DataInput.readShare(): Share {
    val id = readIdentifier()
    val fragment = readByteArray()
    val signature = readByteArray()
    if (id !is ECID)
        throw CodecException("Identifier is not an ECID!", IOException())
    return Share(id, fragment, signature)
}

//Specifies the secret sharing scheme's parameter's including threshold and algorithmparams for the cipher to be used with thes ecret key
class SharedSecretDataProtection(val k: UByte, val params: AlgorithmParameters, vararg val shares: Share)

fun DataOutput.writeSharedSecretDataProtection(protection: SharedSecretDataProtection) {
    writeByte(protection.k.toByte())
    writeByteArray(protection.params.encoded)
    writeByte(protection.shares.size.toUByte().toByte())
    protection.shares.forEach {
        writeShare(it)
    }
}

fun DataInput.readSharedSecretDataProtection(): SharedSecretDataProtection {
    val k = readByte().toUByte()
    val params = AlgorithmParameters.getInstance("AES/GCM/NoPadding").apply { this.init(readByteArray()) }
    val numShares = readByte().toUByte().toInt()
    val shares = Array<Share>(numShares) {
        readShare()
    }
    return SharedSecretDataProtection(k, params, *shares)
}

//basic data container with creator, policy signed by creator and the data portected by this policy
data class DataContainer(
    val creator: ECID,
    val signedPolicy: SignedPolicy,
    val protectedData: ProtectedData
)

fun DataOutput.writeDataContainer(container: DataContainer) {
    writeIdentifier(container.creator)
    writeSignedPolicy(container.signedPolicy)
    writeProtectedData(container.protectedData)
}

fun DataInput.readDataContainer(): DataContainer {
    val id = readIdentifier()
    val pol = readSignedPolicy()
    val protectedData = readProtectedData()
    if (id !is ECID)
        throw CodecException("Identifier is not an ECID!", IOException())
    return DataContainer(id, pol, protectedData)
}

//data container plus signature, making it self-contained s.t. no additional info is required to verify its integrity and origin
data class SignedDataContainer(val dataContainer: DataContainer, val signature: DetachedSignature) {
    companion object {
        fun create(
            creator: ECID,
            signingKey: ECPrivateKey,
            data: ByteArray,
            policy: IdBasedPolicy,
            decryptors: Set<ECID>,
            k: UByte
        ): SignedDataContainer {
            val protectedData = protectDataShared(signingKey, data, decryptors, k)
            val byteOut = ByteArrayOutputStream()
            DataOutputStream(byteOut).apply { writeIdentifier(creator);close() }
            val container = DataContainer(
                creator = creator,
                signedPolicy = SignedPolicy(
                    policy,
                    CommonCrypto.sign(signingKey, policy.toByteArray())
                ),
                protectedData = protectedData
            )
            //not efficient
            val dst = ByteBufferChannel(ByteBuffer.allocateDirect(4096))
            dst.writeDataContainer(container)
            val posAfter = dst.buffer.position()
            dst.buffer.position(0)
            val sig = CommonCrypto.createSignature(signingKey)
            val signature = sig.sign(dst.buffer.slice().apply { limit(posAfter) })
            dst.buffer.position(posAfter)
            return SignedDataContainer(container, signature)
        }
    }
}
fun DataOutput.writeSignedDataContainer(container: SignedDataContainer) {
    writeDataContainer(container.dataContainer)
    writeByteArray(container.signature)
}

fun DataInput.readSignedDataContainer(): SignedDataContainer {
    val dataContainer = readDataContainer()
    val signature = readByteArray()
    return SignedDataContainer(dataContainer, signature)
}

//associates encrypted data, the key used to encrypt it and algorithm parameters, only used internally
private data class EncryptedData(val ciphertext: ByteArray, val key: SecretKey, val params: AlgorithmParameters)

/**
 * generate secret key to AES/GCM-encrypt the encryptedData and return cipher text, secret key and algorithm params for later decryption
 */
private fun encryptData(data: ByteArray): EncryptedData {
    val secretKey = KeyGenerator.getInstance("AES").generateKey()
    val cipher = Cipher.getInstance("AES/GCM/NoPadding")
    cipher.init(Cipher.ENCRYPT_MODE, secretKey)
    val ct = cipher.doFinal(data)
    val params = cipher.parameters
    return EncryptedData(ct, secretKey, params)
}

/**
 * split the secret key amongst the selected identifers, ECIES-encrypted.
 * the plain shares are hashed only for the accessor to check if some decryptor messed up
 */
private fun splitKey(
    signingKey: ECPrivateKey,
    key: SecretKey,
    params: AlgorithmParameters,
    evaluators: Set<ECID>,
    k: UByte
): SharedSecretDataProtection {
    val scheme = Scheme(CommonCrypto.RND, evaluators.size, k.toInt())
    val shares = scheme.split(key.encoded)
    val fragments =
        shares.map {
            Share(
                evaluator = evaluators.elementAt(it.key - 1),
                fragment = CommonCrypto.encrypt(evaluators.elementAt(it.key - 1).publicKey, it.value),
                signature = CommonCrypto.sign(signingKey, it.value)
            )
        }.toTypedArray()
    return SharedSecretDataProtection(k, params, *fragments)
}

/**
 * protect some encrypted Data using secret sharing
 */
fun protectDataShared(
    signingKey: ECPrivateKey,
    data: ByteArray, decryptors: Set<ECID>,
    k: UByte
): ProtectedData {
    val info = encryptData(data)
    return ProtectedData(splitKey(signingKey, info.key, info.params, decryptors, k), info.ciphertext)

}